/*
 * Copyright 1999-2006 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.globus.replica.rls;

import java.util.*;

/**
 * This class implements the interface to an RLS server. The <code>RLSClient</code>
 * depends on the native RLS client libraries, which it links to via JNI.
 * @version 2.0.3
 */
public class RLSClient {

  private int handle;
  private String url;
  private LRC lrc;
  private RLI rli;

  public static final int RLS_SUCCESS = 0;
  public static final int RLS_GLOBUSERR = 1;
  public static final int RLS_INVHANDLE = 2;
  public static final int RLS_BADURL = 3;
  public static final int RLS_NOMEMORY = 4;
  public static final int RLS_OVERFLOW = 5;
  public static final int RLS_BADARG = 6;
  public static final int RLS_PERM = 7;
  public static final int RLS_BADMETHOD = 8;
  public static final int RLS_INVSERVER = 9;
  public static final int RLS_MAPPING_NEXIST = 10;
  public static final int RLS_LFN_EXIST = 11;
  public static final int RLS_LFN_NEXIST = 12;
  public static final int RLS_PFN_EXIST = 13;
  public static final int RLS_PFN_NEXIST = 14;
  public static final int RLS_LRC_EXIST = 15;
  public static final int RLS_LRC_NEXIST = 16;
  public static final int RLS_DBERROR = 17;
  public static final int RLS_RLI_EXIST = 18;
  public static final int RLS_RLI_NEXIST = 19;
  public static final int RLS_MAPPING_EXIST = 20;
  public static final int RLS_INV_ATTR_TYPE = 21;
  public static final int RLS_ATTR_EXIST = 22;
  public static final int RLS_ATTR_NEXIST = 23;
  public static final int RLS_INV_OBJ_TYPE = 24;
  public static final int RLS_INV_ATTR_OP = 25;
  public static final int RLS_UNSUPPORTED = 26;
  public static final int RLS_TIMEOUT = 27;
  public static final int RLS_TOO_MANY_CONNECTIONS = 28;
  public static final int RLS_ATTR_VALUE_NEXIST = 29;
  public static final int RLS_ATTR_INUSE = 30;

  public static final int ADMIN_PING = 0;
  public static final int ADMIN_QUIT = 1;
  public static final int ADMIN_SSU = 2;

  public static final int RLIFLAG_BLOOMFILTER = 0x1;

  /**
   * Connect to an RLS server on the local host at the default port.
   */
  public RLSClient()
	throws RLSException {
    this.url = "rls://localhost";
    handle = connect(this.url);
    lrc = new LRC();
    rli = new RLI();
  }

  /**
   * Connect to an RLS server at the specified url.
   * @param url The location of the RLS server.  It should be of the form
   *   rls://host[:port] or rlsn://host[:port].  The latter form is used
   *   if no authentication is desired.
   */
  public RLSClient(String url)
	throws RLSException {
    this.url = new String(url);
    handle = connect(this.url);
    lrc = new LRC();
    rli = new RLI();
  }

  /**
   * Connect to an RLS server at the specified url, using proxycert for
   * authentication.
   * @param url The location of the RLS server.  It should be of the form
   *   rls://host[:port] or rlsn://host[:port].  The latter form is used
   *   if no authentication is desired.
   * @param proxyfile Proxy certificate filename to use for authentication.
   */
  public RLSClient(String url, String proxyfile)
	throws RLSException {
    proxycertificate(proxyfile);
    this.url = new String(url);
    handle = connect(this.url);
    lrc = new LRC();
    rli = new RLI();
  }

  /**
   * Connect to an RLS server at the specified url, using certfile/key for
   * authentication.
   * @param url The location of the RLS server.  It should be of the form
   *   rls://host[:port] or rlsn://host[:port].  The latter form is used
   *   if no authentication is desired.
   * @param certfile Certificate file to use for authentication.
   * @param keyfile Key file to use for authentication.
   */
  public RLSClient(String url, String certfile, String keyfile)
	throws RLSException {
    certificate(certfile, keyfile);
    this.url = new String(url);
    handle = connect(this.url);
    lrc = new LRC();
    rli = new RLI();
  }

  /**
   * Miscellaneous administrative operations.  Most operations require the
   * ADMIN privilege.
   *
   * @param h Handle connected to RLS server.
   *
   * @param cmd Command to be sent to RLS server.
   *   See {@link org.globus.replica.rls.RLSClient#ADMIN_PING admin commands}.
   */
  public void Admin(int cmd)
	throws RLSException {
    admin(handle, cmd);
  }

  /**
   * Set name of cert and key files.
   */
  public static void Certificate(String certfile, String keyfile) {
    certificate(certfile, keyfile);
  }

  /**
   * Set name of proxy cert file.
   */
  public static void ProxyCertificate(String proxyfile) {
    proxycertificate(proxyfile);
  }

  /**
   * Close connection to RLS server.
   */
  public void Close()
	throws RLSException {
    close(handle);
  }

  /**
   * Map status code returned by RLS to string.
   *
   * @param rc Status code.
   *
   * @return String Error message.
   */
  public String getErrorMessage(int rc) {
    return geterrormessage(rc);
  }

  /**
   * Returns an instance of the LRC.
   *
   * @return Reference to LRC object.
   */
  public LRC getLRC() {
    return lrc;
  }

  /**
   * Returns an instance of the RLI.
   *
   * @return Reference to RLI object.
   */
  public RLI getRLI() {
    return rli;
  }

  /**
   * Get timeout, in seconds, used in IO calls.  The default is 30 seconds.
   *
   * @return seconds Timeout value.
   */
  public int GetTimeout() {
    return gettimeout();
  }

  /**
   * Set timeout value for IO calls.  This value applies to all RLS
   * connections, not just the current one.
   *
   * @param seconds Timeout value.
   */
  public void SetTimeout(int seconds) {
    settimeout(seconds);
  }

  /**
   * Return URL of RLS server.
   */
  public String GetURL() {
    return url;
  }

  /**
   * Add an attribute to an object in the LRC database.
   *
   * @param key Logical or Physical File Name (LFN or PFN) that identifies
   *   object attribute should be added to.
   *
   * @param attr Attribute to be added to object.  See
   *   {@link org.globus.replica.rls.RLSAttribute#LRC_LFN object types}.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#attributeAdd
   *   getLRC().attributeAdd()}
   */
  public void LRCAttrAdd(String key, RLSAttribute attr)
	throws RLSException {
    lrc.attributeAdd(key, attr);
  }

  /**
   * Define new attribute in LRC database.
   *
   * @param name Name of attribute.
   *
   * @param objtype Object (LFN or PFN) type that attribute applies to.  See
   *   {@link org.globus.replica.rls.RLSAttribute#LRC_LFN object types}.
   *
   * @param valtype Type of attribute value.  See
   *   {@link org.globus.replica.rls.RLSAttribute#DATE attribute types}.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#attributeCreate
   *   getLRC().attributeCreate()}
   */
  public void LRCAttrCreate(String name, int objtype, int valtype)
	throws RLSException {
    lrc.attributeCreate(name, objtype, valtype);
  }

  /**
   * Undefine attribute in LRC database.
   *
   * @param name Name of attribute.
   *
   * @param objtype Object (LFN or PFN) type that attribute applies to.  See
   *   {@link org.globus.replica.rls.RLSAttribute#LRC_LFN object types}.
   *
   * @param clearvalues  If true then any any values for this
   *   attribute are first removed from the objects they're associated with.
   *   If false and any values exist then an exception is thrown.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#attributeDelete
   *   getLRC().attributeDelete()}
   */
  public void LRCAttrDelete(String name, int objtype, boolean clearvalues)
	throws RLSException {
    lrc.attributeDelete(name, objtype, clearvalues);
  }

  /**
   * Return definitions of attributes in LRC database.
   *
   * @param name Name of attribute.  If name is null all attributes of the
   *   specified objtype are returned.
   *
   * @param objtype Object (LFN or PFN) type that attribute applies to.
   *   See {@link org.globus.replica.rls.RLSAttribute#LRC_LFN object types}.
   *
   * @return attr_list Any attribute definitions found will be returned as a
   *   list of {@link org.globus.replica.rls.RLSAttribute RLSAttribute} objects.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#attributeGet
   *   getLRC().attributeGet()}
   */
  public ArrayList LRCAttrGet(String name, int objtype)
	throws RLSException {
    return lrc.attributeGet(name, objtype);
  }

  /**
   * Modify an attribute value.
   *
   * @param key Name of object (LFN or PFN).
   *
   * @param attr Attribute to be modified.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#attributeModify
   *   getLRC().attributeModify()}
   */
  public void LRCAttrModify(String key, RLSAttribute attr)
	throws RLSException {
    lrc.attributeModify(key, attr);
  }

  /**
   * Remove an attribute from an object (LFN or PFN) in the LRC database.
   *
   * @param key Name of object (LFN or PFN).
   *
   * @param attr Attribute to be removed.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#attributeRemove
   *   getLRC().attributeRemove()}
   */
  public void LRCAttrRemove(String key, RLSAttribute attr)
	throws RLSException {
    lrc.attributeRemove(key, attr);
  }

  /**
   * Search for objects (LFNs or PFNs) in a LRC database that have
   * the specified attribute whose value matches a boolean expression.
   *
   * @param name  Name of attribute.
   *
   * @param objtype Object (LFN or PFN) type that attribute applies to.
   *
   * @param op Operator to be used in searching for values.  See
   *   {@link org.globus.replica.rls.RLSAttribute#OPALL operators}.
   *
   * @param operand1 First operand in boolean expression.
   *
   * @param operand2 Second operand in boolean expression, only used when op is
   *	{@link org.globus.replica.rls.RLSAttribute#OPBTW RLSAttribute.OPBTW}.
   *
   * @param offlim Offset and limit used to retrieve results incrementally.
   *	Use null, or 0,0, to retrieve all results.
   *
   * @return attr_obj_list Any objects with the specified attribute will be
   *    returned, with the attribute value, in a list of
   *    {@link org.globus.replica.rls.RLSAttributeObject RLSAttributeObject}
   *    objects.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#attributeSearch
   *   getLRC().attributeSearch()}
   */
  public ArrayList LRCAttrSearch(String name, int objtype, int op,
				 RLSAttribute op1, RLSAttribute op2,
				 RLSOffsetLimit offlim)
	throws RLSException {
    return lrc.attributeSearch(name, objtype, op, op1, op2, offlim);
  }

  /**
   * Return attributes in LRC database for specified object (LFN or PFN).
   *
   * @param key Logical or Physical File Name (LFN or PFN) that identifies
   *   object attributes should be retrieved for.
   *
   * @param name Name of attribute to retrieve.  If null all attributes for
   *	key, objtype are returned.
   *
   * @param objtype Object (LFN or PFN) type that attribute applies to.
   *
   * @return attr_list Any attributes found will be returned in a list
   *	of {@link org.globus.replica.rls.RLSAttribute RLSAttribute} objects.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#attributeValueGet
   *   getLRC().attributeValueGet()}
   */
  public ArrayList LRCAttrValueGet(String key, String name, int objtype)
	throws RLSException {
    return lrc.attributeValueGet(key, name, objtype);
  }

  /**
   * Add mapping to PFN to an existing LFN.
   *
   * @param lfn LFN to add pfn mapping to, should already exist.
   *
   * @param pfn PFN that lfn should map to.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#add
   *   getLRC().add()}
   */
  public void LRCAdd(String lfn, String pfn)
	throws RLSException {
    lrc.add(lfn, pfn);
  }

  /**
   * Remove all lfns, pfns, mappings and attribute values from LRC database.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#clear
   *   getLRC().clear()}
   */
  public void LRCClear() throws RLSException {
    lrc.clear();
  }

  /**
   * Create new mapping to PFN from LFN.
   *
   * @param lfn LFN to add pfn mapping to, should not exist.
   *
   * @param pfn PFN that lfn should map to.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#create
   *   getLRC().create()}
   */
  public void LRCCreate(String lfn, String pfn)	throws RLSException {
    lrc.create(lfn, pfn);
  }

  /**
   * Delete mapping from LRC database.
   *
   * @param lfn LFN to remove mapping from.
   *
   * @param pfn PFN that LFN maps to be removed.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#delete
   *   getLRC().delete()}
   */
  public void LRCDelete(String lfn, String pfn)
	throws RLSException {
    lrc.delete(lfn, pfn);
  }

  /**
   * Test if object (LFN or PFN) exists.
   *
   * @param key Name of object.
   *
   * @param objtype Type of object.
   *
   * @return boolean True if object exists, else false.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#exists
   *   getLRC().exists()}
   */
  public boolean LRCExists(String key, int objtype)
	throws RLSException {
    return lrc.exists(key, objtype);
  }

  /**
   * Return LFNs mapped to PFN in the LRC database.
   *
   * @param pfn PFN to search for.
   *
   * @param offlim Offset and limit used to retrieve results incrementally.
   *	Use null, or 0,0, to retrieve all results.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2
   *    RLSString2} objects containing matching lfn,pfn mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#getLFN
   *   getLRC().getLFN()}
   */
  public ArrayList LRCGetLFN(String pfn, RLSOffsetLimit offlim)
	throws RLSException {
    return lrc.getLFN(pfn, offlim);
  }

  /**
   * Return all LFNs mapped to PFN in the LRC database.
   *
   * @param pfn PFN to search for.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2
   *    rls.RLSString2} objects containing matching lfn,pfn mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#getLFN
   *   getLRC().getLFN()}
   */
  public ArrayList LRCGetLFN(String pfn)
	throws RLSException {
    return lrc.getLFN(pfn);
  }

  /**
   * Return LFNs mapped to PFNs matching pfnpat in the LRC database.
   *
   * @param pfnpat PFN pattern to search for.
   *
   * @param offlim Offset and limit used to retrieve results incrementally.
   *	Use null, or 0,0, to retrieve all results.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
   *    objects containing matching lfn,pfn mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#getLFNWC
   *   getLRC().getLFNWC()}
   */
  public ArrayList LRCGetLFNWC(String pfnpat, RLSOffsetLimit offlim)
	throws RLSException {
    return lrc.getLFNWC(pfnpat, offlim);
  }

  /**
   * Return all LFNs mapped to PFNs matching pfnpat in the LRC database.
   *
   * @param pfnpat PFN pattern to search for.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
   *    objects containing matching lfn,pfn mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#getLFNWC
   *   getLRC().getLFNWC()}
   */
  public ArrayList LRCGetLFNWC(String pfnpat)
	throws RLSException {
    return lrc.getLFNWC(pfnpat);
  }

  /**
   * Return PFNs mapped to LFN in the LRC database.
   *
   * @param lfn LFN to search for.
   *
   * @param offlim Offset and limit used to retrieve results incrementally.
   *	Use null, or 0,0, to retrieve all results.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2} objects
   *    containing matching lfn,pfn mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#getPFN
   *   getLRC().gePLFN()}
   */
  public ArrayList LRCGetPFN(String lfn, RLSOffsetLimit offlim)
	throws RLSException {
    return lrc.getPFN(lfn, offlim);
  }

  /**
   * Return all PFNs mapped to LFN in the LRC database.
   *
   * @param lfn LFN to search for.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
   *    objects containing matching lfn,pfn mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#getPFN
   *   getLRC().gePLFN()}
   */
  public ArrayList LRCGetPFN(String lfn)
	throws RLSException {
    return lrc.getPFN(lfn);
  }

  /**
   * Return PFNs mapped to LFNs matching lfnpat in the LRC database.
   *
   * @param lfnpat LFN pattern to search for.
   *
   * @param offlim Offset and limit used to retrieve results incrementally.
   *	Use null, or 0,0, to retrieve all results.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
   *    objects containing matching lfn,pfn mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#getPFNWC
   *   getLRC().gePLFNWC()}
   */
  public ArrayList LRCGetPFNWC(String lfnpat, RLSOffsetLimit offlim)
	throws RLSException {
    return lrc.getPFNWC(lfnpat, offlim);
  }

  /**
   * Return all PFNs mapped to LFNs matching lfnpat in the LRC database.
   *
   * @param lfnpat LFN pattern to search for.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
   *    objects containing matching lfn,pfn mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#getPFNWC
   *   getLRC().gePLFNWC()}
   */
  public ArrayList LRCGetPFNWC(String lfnpat)
	throws RLSException {
    return lrc.getPFNWC(lfnpat);
  }

  /**
   * Test if mapping exists in LRC.
   *
   * @param lfn Name of LFN.
   *
   * @param pfn Name of PFN.
   *
   * @return boolean True if mapping exists, else false.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#mappingExists
   *   getLRC().mappingExists()}
   */
  public boolean LRCMappingExists(String lfn, String pfn)
	throws RLSException {
    return lrc.mappingExists(lfn, pfn);
  }

  /**
   * LRC servers send information about LFNs in their database to the
   * the list of RLI servers in the database, added with the following
   * function.  Updates may be partitioned amongst multiple RLIs by
   * specifying one or more patterns for an RLI.
   *
   * @param rliurl URL of RLI server that LRC should send updates to.
   *
   * @param flags Should be zero or
   *    {@link org.globus.replica.rls.RLSClient#RLIFLAG_BLOOMFILTER RLSClient.RLIFLAG_BLOOMFILTER}
   *    if the RLI should be updated via Bloom filters.
   *
   * @param pattern If not NULL used to filter which LFNs are sent to rli_url.
   *	Standard Unix wildcard characters (*, ?) may be used to do wildcard
   *	matches.  Patterns are ignored if Bloom filters are used for updates.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#rliAdd
   *   getLRC().rliAdd()}
   */
  public void LRCRLIAdd(String rliurl, int flags, String pattern)
	throws RLSException {
    lrc.rliAdd(rliurl, flags, pattern);
  }

  /**
   * Delete an entry from the LRC rli/partition tables.
   *
   * @param rliurl URL of RLI server to remove from LRC partition table.
   *
   * @param pattern If not null then only the specific rli_url/pattern is
   * removed, else all partition information for rli_url is removed.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#rliDelete
   *   getLRC().rliDelete()}
   */
  public void LRCRLIDelete(String rliurl, String pattern)
	throws RLSException {
    lrc.rliDelete(rliurl, pattern);
  }

  /**
   * Get RLI update partitions from LRC server.
   *
   * @param rliurl If not NULL identifies RLI that partition data will be
   *   retrieved for.  If null then all RLIs are retrieved.
   *
   * @param pattern If not NULL returns only partitions with matching patterns,
   *	otherwise all patterns are retrieved.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2} objects,
   *    containing the RLI URLs and patterns (if any) updated by the
   *    LRC.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#rliGetPart
   *   getLRC().rliGetPart()}
   */
  public ArrayList LRCRLIGetPart(String rliurl, String pattern)
	throws RLSException {
    return lrc.rliGetPart(rliurl, pattern);
  }

  /**
   * Get info about RLI server updated by an LRC server.
   *
   * @param rliurl URL of RLI server to retrieve info for.
   *
   * @return info Data about the RLI server will be returned in an
   *   {@link org.globus.replica.rls.RLSRLIInfo RLSRLIInfo} object.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#rliInfo
   *   getLRC().rliInfo()}
   */
  public RLSRLIInfo LRCRLIInfo(String rliurl)
	throws RLSException {
    return lrc.rliInfo(rliurl);
  }

  /**
   * Return URLs of RLIs that LRC sends updates to.
   *
   * @return A list of RLSRLIInfo objects containing the RLI URLs updated by
   *   this LRC.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.LRC#rliList
   *   getLRC().rliList()}
   */
  public ArrayList LRCRLIList()
	throws RLSException {
    return lrc.rliList();
  }

  /**
   * Return LRCs mapped to LFN in the LRC database.
   *
   * @param lfn LFN to search for.
   *
   * @param offlim Offset and limit used to retrieve results incrementally.
   *	Use null, or 0,0, to retrieve all results.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
   *    objects containing matching lfn,lrc mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.RLI#getLRC
   *   getRLI().getLRC()}
   */
  public ArrayList RLIGetLRC(String lfn, RLSOffsetLimit offlim)
	throws RLSException {
    return rli.getLRC(lfn, offlim);
  }

  /**
   * Return all LRCs mapped to LFNs matching lfnpat in the LRC database.
   *
   * @param lfnpat LFN pattern to search for.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
   *    objects containing matching lfn,lrc mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.RLI#getLRC
   *   getRLI().getLRC()}
   */
  public ArrayList RLIGetLRC(String lfn)
	throws RLSException {
    return rli.getLRC(lfn);
  }

  /**
   * Return LRCs mapped to LFNs matching lfnpat in the LRC database.
   *
   * @param lfnpat LFN pattern to search for.
   *
   * @param offlim Offset and limit used to retrieve results incrementally.
   *	Use null, or 0,0, to retrieve all results.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2} objects
   *    containing matching lfn,lrc mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.RLI#getLRCWC
   *   getRLI().getLRCWC()}
   */
  public ArrayList RLIGetLRCWC(String lfnpat, RLSOffsetLimit offlim)
	throws RLSException {
    return rli.getLRCWC(lfnpat, offlim);
  }

  /**
   * Return all LRCs mapped to lfns matching lfnpat in the LRC database.
   *
   * @param lfnpat LFN pattern to search for.
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2} objects
   *    containing matching lfn,lrc mappings.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.RLI#getLRCWC
   *   getRLI().getLRCWC()}
   */
  public ArrayList RLIGetLRCWC(String lfnpat)
	throws RLSException {
    return rli.getLRCWC(lfnpat);
  }

  /**
   * Test if object (LFN or LRC) exists.
   *
   * @param key Name of object.
   *
   * @param objtype Type of object.
   *
   * @return boolean True if object exists, else false.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.RLI#exists
   *   getRLI().exists()}
   */
  public boolean RLIExists(String key, int objtype)
	throws RLSException {
    return rli.exists(key, objtype);
  }

  /**
   * Return URLs of LRCs that update this RLI.
   *
   * @return A list of RLSLRCInfo objects containing the RLI URLs updated by
   *   this LRC.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.RLI#lrcList
   *   getRLI().lrcList()}
   */
  public ArrayList RLILRCList()
	throws RLSException {
    return rli.lrcList();
  }

  /**
   * Test if mapping exists in RLI.
   *
   * @param lfn Name of LFN.
   *
   * @param lrc Name of LRC.
   *
   * @return boolean True if mapping exists, else false.
   *
   * @deprecated Use {@link org.globus.replica.rls.RLSClient.RLI#mappingExists
   *   getRLI().mappingExists()}
   */
  public boolean RLIMappingExists(String lfn, String lrc)
	throws RLSException {
    return rli.mappingExists(lfn, lrc);
  }

  /**
   * Return server configuration.
   *
   * @param option Configuration option to retreive.  If null all
   *   options are returned
   *
   * @return str2_list List of {@link org.globus.replica.rls.RLSString2
   *  RLSString2} objects containing configuration option names and values.
   */
  public ArrayList GetConfiguration(String option)
	throws RLSException {
    return getconfiguration(handle, option);
  }

  /**
   * Set server configuration.
   *
   * @param option Configuration option to set.
   *
   * @param value Value to set.
   */
  public int SetConfiguration(String option, String value)
	throws RLSException {
    return setconfiguration(handle, option, value);
  }

  /**
   * Return stats about an RLS server.
   *
   * @return RLSStats.
   *
   */
  public RLSStats Stats()
	throws RLSException {
    return stats(handle);
  }

  /**
   * Make call to system(), hack to allow me to invoke globus-url-copy for
   * the sc2002 demo.
   *
   * @return int system() result.
   */
  static public int OSSystem(String cmd) {
    return system(cmd);
  }

  static {
    String flavors[] = {
      "gcc32dbgpthr",
      "gcc32dbg",
      "gcc32pthr",
      "gcc32",
    };

    String flavor = System.getProperty("globus.flavor");
    if (flavor != null)
      System.loadLibrary("globus_rls_client_jni_" + flavor);
    else {
      for (int i = 0; i < flavors.length; i++) {
	try {
	  System.loadLibrary("globus_rls_client_jni_" + flavors[i]);
	  break;
	} catch (UnsatisfiedLinkError le) {
	  System.out.println("Couldn't find flavor " + flavors[i]);
	}
      }
    }
    init();
  }

  private native void admin(int handle, int cmd);
  private native int connect(String url);
  private native void close(int handle);
  private static native void certificate(String certfile, String keyfile);
  private static native void proxycertificate(String proxyfile);
  private native String geterrormessage(int rc);
  private native int gettimeout();
  private native void settimeout(int seconds);

  private native void lrcattradd(int handle, String key, RLSAttribute attr);
  private native ArrayList lrcattraddbulk(int handle, int attrobjlist);
  private native void lrcattrcreate(int handle, String name, int objtype,
				    int valtype);
  private native void lrcattrdelete(int handle, String name, int objtype,
				    int clearvalues);
  private native ArrayList lrcattrget(int handle, String name, int objtype);
  private native void lrcattrmodify(int handle, String key,RLSAttribute attr);
  private native void lrcattrremove(int handle, String key,RLSAttribute attr);
  private native ArrayList lrcattrremovebulk(int handle, int attrobjlist);
  private native ArrayList lrcattrsearch(int handle, String name, int objtype,
				    int op, RLSAttribute op1, RLSAttribute op2,
				    RLSOffsetLimit offlim);
  private native ArrayList lrcattrvalueget(int handle, String key, String name,
					   int objtype);
  private native ArrayList lrcattrvaluegetbulk(int handle, int keylist,
					       String name, int objtype);
  private native void lrcadd(int handle, String lfn, String pfn);
  private native ArrayList lrcaddbulk(int handle, int str2);
  private native void lrcclear(int handle);
  private native void lrccreate(int handle, String lfn, String pfn);
  private native ArrayList lrccreatebulk(int handle, int str2);
  private native void lrcdelete(int handle, String lfn, String pfn);
  private native ArrayList lrcdeletebulk(int handle, int str2);
  private native boolean lrcexists(int handle, String key, int objtype);
  private native ArrayList lrcexistsbulk(int handle, int strlist, int objtype);
  private native ArrayList lrcgetlfn(int handle, String pfn,
				     RLSOffsetLimit jofflim);
  private native ArrayList lrcgetlfnbulk(int handle, int strlist);
  private native ArrayList lrcgetlfnwc(int handle, String pfnpat,
				       RLSOffsetLimit jofflim);
  private native ArrayList lrcgetpfn(int handle, String lfn,
				     RLSOffsetLimit offlim);
  private native ArrayList lrcgetpfnbulk(int handle, int strlist);
  private native ArrayList lrcgetpfnwc(int handle, String lfnpat,
				       RLSOffsetLimit offlim);
  private native boolean lrcmappingexists(int handle, String lfn, String pfn);
  private native void lrcrenamelfn(int handle, String oldname, String newname);
  private native ArrayList lrcrenamelfnbulk(int handle, int str2);
  private native void lrcrenamepfn(int handle, String oldname, String newname);
  private native ArrayList lrcrenamepfnbulk(int handle, int str2);
  private native void lrcrliadd(int handle, String rliurl, int flags,
				String pattern);
  private native void lrcrlidelete(int handle, String rliurl, String pattern);
  private native ArrayList lrcrligetpart(int handle, String rliurl, String pattern);
  private native RLSRLIInfo lrcrliinfo(int handle, String rliurl);
  private native ArrayList lrcrlilist(int handle);
  private native int makestrlist(int str2list, String lfn);
  private native int makestr2list(int str2list, String lfn, String pfn);
  private native int makeaolist(int attrobjlist,String key,RLSAttribute attr);
  private native ArrayList rligetlrc(int handle, String lfn,
				     RLSOffsetLimit offlim);
  private native ArrayList rligetlrcbulk(int handle, int strlist);
  private native ArrayList rligetlrcwc(int handle, String lfnpat,
				       RLSOffsetLimit offlim);
  private native boolean rliexists(int handle, String key, int objtype);
  private native ArrayList rliexistsbulk(int handle, int strlist, int objtype);
  private native boolean rlimappingexists(int handle, String lfn, String lrc);
  private native ArrayList rlirlilist(int handle);
  private native ArrayList rlisenderlist(int handle);
  private native ArrayList getconfiguration(int handle, String option);
  private native int setconfiguration(int handle, String option, String value);
  private native RLSStats stats(int handle);
  private static native int system(String cmd);

  static native void init();



  /**
   * This class implements the LRC interface to an RLS server.
   * @version 2.0.3
   */
  public class LRC {

    /**
     * Add mapping to PFN to an existing LFN.
     *
     * @param lfn LFN to add pfn mapping to, should already exist.
     *
     * @param pfn PFN that lfn should map to.
     */
    public void add(String lfn, String pfn)
	  throws RLSException {
      lrcadd(handle, lfn, pfn);
    }

    /**
     * Bulk add mappings to PFNs to existing LFNs.
     *
     * @param str2list <code>ArrayList</code> of <code>RLSString2</code>
     *   elements, where <code>s1</code> is the LFN to add PFN mapping to,
     *   and <code>s2</code> is the PFN that LFN should map to. The LFN
     *   should already exist.
     *
     * @return List of failed updates.  ArrayList elements will be of
     *   type <code>RLSString2Bulk</code>.
     */
    public ArrayList addBulk(ArrayList str2list) throws RLSException {
      int p = 0;
      
      Iterator iter = str2list.iterator();
      while (iter.hasNext()) {
	RLSString2 str2 = (RLSString2) iter.next();
	p = makestr2list(p, str2.s1, str2.s2);
      }
      return lrcaddbulk(handle, p);
    }

    /**
     * Add an attribute to an object in the LRC database.
     *
     * @param key Logical or Physical File Name (LFN or PFN) that identifies
     *   object attribute should be added to.
     *
     * @param attr Attribute to be added to object.  See
     *   {@link org.globus.replica.rls.RLSAttribute#LRC_LFN object types}.
     */
    public void attributeAdd(String key, RLSAttribute attr)
	  throws RLSException {
      lrcattradd(handle, key, attr);
    }

    /**
     * Bulk add attributes.
     *
     * @param attr_obj_list Object and attributes to add to them. Elements
     *   must be of type <code>RLSAttributeObject</code>.
     *
     * @return List of failed updates.  ArrayList elements will be of
     *   type RLSString2Bulk.
     */
    public ArrayList attributeAddBulk(ArrayList attr_obj_list) throws RLSException {
      int p = 0;
      
      Iterator iter = attr_obj_list.iterator();
      while (iter.hasNext()) {
	RLSAttributeObject ao = (RLSAttributeObject) iter.next();
	p = makeaolist(p, ao.key, ao.attr);
      }
      return lrcattraddbulk(handle, p);
    }

    /**
     * Define new attribute in LRC database.
     *
     * @param name Name of attribute.
     *
     * @param objtype Object (LFN or PFN) type that attribute applies to.  See
     *   {@link org.globus.replica.rls.RLSAttribute#LRC_LFN object types}.
     *
     * @param valtype Type of attribute value.  See
     *   {@link org.globus.replica.rls.RLSAttribute#DATE attribute types}.
     */
    public void attributeCreate(String name, int objtype, int valtype)
	  throws RLSException {
      lrcattrcreate(handle, name, objtype, valtype);
    }

    /**
     * Undefine attribute in LRC database.
     *
     * @param name Name of attribute.
     *
     * @param objtype Object (LFN or PFN) type that attribute applies to.  See
     *   {@link org.globus.replica.rls.RLSAttribute#LRC_LFN object types}.
     *
     * @param clearvalues  If true then any any values for this
     *   attribute are first removed from the objects they're associated with.
     *   If false and any values exist then an exception is thrown.
     */
    public void attributeDelete(String name, int objtype, boolean clearvalues)
	  throws RLSException {
      lrcattrdelete(handle, name, objtype, clearvalues ? 1 : 0);
    }

    /**
     * Return definitions of attributes in LRC database.
     *
     * @param name Name of attribute.  If name is null all attributes of the
     *   specified objtype are returned.
     *
     * @param objtype Object (LFN or PFN) type that attribute applies to.
     *   See {@link org.globus.replica.rls.RLSAttribute#LRC_LFN object types}.
     *
     * @return attr_list Any attribute definitions found will be returned as a
     *   list of {@link org.globus.replica.rls.RLSAttribute RLSAttribute}
     *   objects.
     */
    public ArrayList attributeGet(String name, int objtype)
	  throws RLSException {
      return lrcattrget(handle, name, objtype);
    }

    /**
     * Modify an attribute value.
     *
     * @param key Name of object (LFN or PFN).
     *
     * @param attr Attribute to be modified.
     */
    public void attributeModify(String key, RLSAttribute attr)
	  throws RLSException {
      lrcattrmodify(handle, key, attr);
    }

    /**
     * Remove an attribute from an object (LFN or PFN) in the LRC database.
     *
     * @param key Name of object (LFN or PFN).
     *
     * @param attr Attribute to be removed.
     */
    public void attributeRemove(String key, RLSAttribute attr)
	  throws RLSException {
      lrcattrremove(handle, key, attr);
    }

    /**
     * Bulk remove attributes.
     *
     * @param attr_obj_list Object and attributes to remove from them.
     *   Elements must be of type <code>RLSAttributeObject<code>.
     *
     * @return List of failed updates.  ArrayList elements will be of
     *   type RLSString2Bulk.
     */
    public ArrayList attributeRemoveBulk(ArrayList attr_obj_list) throws RLSException {
      int p = 0;
      
      Iterator iter = attr_obj_list.iterator();
      while (iter.hasNext()) {
	RLSAttributeObject ao = (RLSAttributeObject) iter.next();
	p = makeaolist(p, ao.key, ao.attr);
      }
      return lrcattrremovebulk(handle, p);
    }

    /**
     * Search for objects (LFNs or PFNs) in a LRC database that have
     * the specified attribute whose value matches a boolean expression.
     *
     * @param name  Name of attribute.
     *
     * @param objtype Object (LFN or PFN) type that attribute applies to.
     *
     * @param op Operator to be used in searching for values.  See
     *   {@link org.globus.replica.rls.RLSAttribute#OPALL operators}.
     *
     * @param operand1 First operand in boolean expression.
     *
     * @param operand2 Second operand in boolean expression, only used when op is
     *	{@link org.globus.replica.rls.RLSAttribute#OPBTW RLSAttribute.OPBTW}.
     *
     * @param offlim Offset and limit used to retrieve results incrementally.
     *	Use null, or 0,0, to retrieve all results.
     *
     * @return attr_obj_list Any objects with the specified attribute will be
     *    returned, with the attribute value, in a list of
     *    {@link org.globus.replica.rls.RLSAttributeObject RLSAttributeObject}
     *    objects.
     */
    public ArrayList attributeSearch(String name, int objtype, int op,
				     RLSAttribute op1, RLSAttribute op2,
				     RLSOffsetLimit offlim)
	  throws RLSException {
      return lrcattrsearch(handle, name, objtype, op, op1, op2, offlim);
    }

    /**
     * Return attributes in LRC database for specified object (LFN or PFN).
     *
     * @param key Logical or Physical File Name (LFN or PFN) that identifies
     *   object attributes should be retrieved for.
     *
     * @param name Name of attribute to retrieve.  If null all attributes for
     *	key, objtype are returned.
     *
     * @param objtype Object (LFN or PFN) type that attribute applies to.
     *
     * @return attr_list Any attributes found will be returned in a list
     *	of {@link org.globus.replica.rls.RLSAttribute RLSAttribute} objects.
     */
    public ArrayList attributeValueGet(String key, String name, int objtype)
	  throws RLSException {
      return lrcattrvalueget(handle, key, name, objtype);
    }

    /**
     * Bulk return attributes in LRC database for specified objects (LFNs or
     * PFNs).
     *
     * @param keylist Logical or Physical File Name (LFN or PFN) that identify
     *   objects that attributes should be retrieved for. Elements of type
     *   <code>String</code>.
     *
     * @param name Name of attribute to retrieve.  If null all attributes for
     *	key, objtype are returned.
     *
     * @param objtype Object (LFN or PFN) type that attribute applies to.
     *
     * @return attr_obj_list Any attributes found will be returned in a list
     *	of {@link org.globus.replica.rls.RLSAttributeObject
     *  RLSAttributeObject} objects.
     */
    public ArrayList attributeValueGetBulk(ArrayList keylist, String name,
					   int objtype)
	  throws RLSException {
      int p = 0;

      Iterator iter = keylist.iterator();
      while (iter.hasNext()) {
	String s = (String) iter.next();
	p = makestrlist(p, s);
      }
      return lrcattrvaluegetbulk(handle, p, name, objtype);
    }

    /**
     * Remove all lfns, pfns, mappings and attribute values from LRC database.
     */
    public void clear() throws RLSException {
      lrcclear(handle);
    }

    /**
     * Create new mapping to PFN from LFN.
     *
     * @param lfn LFN to add pfn mapping to, should not exist.
     *
     * @param pfn PFN that lfn should map to.
     */
    public void create(String lfn, String pfn)
	  throws RLSException {
      lrccreate(handle, lfn, pfn);
    }

    /**
     * Bulk create LFN,PFN mappings.
     *
     * @param str2list List of <code>RLSString2</code> elements, where
     *   <code>s1</code> is the LFN and <code>s2</code> is the PFN.
     *
     * @return List of failed updates.  ArrayList elements will be of
     *   type RLSString2Bulk.
     */
    public ArrayList createBulk(ArrayList str2list) throws RLSException {
      int p = 0;
      
      Iterator iter = str2list.iterator();
      while (iter.hasNext()) {
	RLSString2 str2 = (RLSString2) iter.next();
	p = makestr2list(p, str2.s1, str2.s2);
      }
      return lrccreatebulk(handle, p);
    }

    /**
     * Delete mapping from LRC database.
     *
     * @param lfn LFN to remove mapping from.
     *
     * @param pfn PFN that LFN maps to be removed.
     */
    public void delete(String lfn, String pfn)
	  throws RLSException {
      lrcdelete(handle, lfn, pfn);
    }

    /**
     * Bulk delete LFN,PFN mappings.
     *
     * @param str2list List of <code>RLSString2</code> elements, where
     *   <code>s1</code> is the LFN and <code>s2</code> is the PFN.
     *
     * @return List of failed updates.  ArrayList elements will be of
     *   type RLSString2Bulk.
     */
    public ArrayList deleteBulk(ArrayList str2list) throws RLSException {
      int p = 0;
      
      Iterator iter = str2list.iterator();
      while (iter.hasNext()) {
	RLSString2 str2 = (RLSString2) iter.next();
	p = makestr2list(p, str2.s1, str2.s2);
      }
      return lrcdeletebulk(handle, p);
    }

    /**
     * Test if object (LFN or PFN) exists.
     *
     * @param key Name of object.
     *
     * @param objtype Type of object.
     *
     * @return boolean True if object exists, else false.
     */
    public boolean exists(String key, int objtype)
	  throws RLSException {
      return lrcexists(handle, key, objtype);
    }

    /**
     * Bulk test if object (LFN or PFN) exists.
     *
     * @param strlist List of objects (LFN or PFN) to test for. Elements
     *   are of type <code>String </code>.
     *
     * @param objtype Type of object.
     *
     * @return List of query results.  ArrayList elements will be of
     *   type RLSString2Bulk.
     */
    public ArrayList existsBulk(ArrayList strlist, int objtype) throws RLSException {
      int p = 0;
      
      Iterator iter = strlist.iterator();
      while (iter.hasNext())
	p = makestrlist(p, (String) iter.next());
      return lrcexistsbulk(handle, p, objtype);
    }
    
    /**
     * Return LFNs mapped to PFN in the LRC database.
     *
     * @param pfn PFN to search for.
     *
     * @param offlim Offset and limit used to retrieve results incrementally.
     *	Use null, or 0,0, to retrieve all results.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2
     *    RLSString2} objects containing matching lfn,pfn mappings.
     */
    public ArrayList getLFN(String pfn, RLSOffsetLimit offlim)
	  throws RLSException {
      return lrcgetlfn(handle, pfn, offlim);
    }

    /**
     * Return all LFNs mapped to PFN in the LRC database.
     *
     * @param pfn PFN to search for.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2
     *    rls.RLSString2} objects containing matching lfn,pfn mappings.
     */
    public ArrayList getLFN(String pfn)
	  throws RLSException {
      return lrcgetlfn(handle, pfn, null);
    }

    /**
     * Bulk query of PFNs.
     *
     * @param strlist List of PFNs to search for. Elements of type
     *   <code>String</code>.
     *
     * @return List of query results.  ArrayList elements will be of
     *   type RLSString2Bulk.
     */
    public ArrayList getLFNBulk(ArrayList strlist) throws RLSException {
      int p = 0;
      
      Iterator iter = strlist.iterator();
      while (iter.hasNext())
	p = makestrlist(p, (String) iter.next());
      return lrcgetlfnbulk(handle, p);
    }

    /**
     * Return LFNs mapped to PFNs matching pfnpat in the LRC database.
     *
     * @param pfnpat PFN pattern to search for.
     *
     * @param offlim Offset and limit used to retrieve results incrementally.
     *	Use null, or 0,0, to retrieve all results.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
     *    objects containing matching lfn,pfn mappings.
     */
    public ArrayList getLFNWC(String pfnpat, RLSOffsetLimit offlim)
	  throws RLSException {
      return lrcgetlfnwc(handle, pfnpat, offlim);
    }

    /**
     * Return all LFNs mapped to PFNs matching pfnpat in the LRC database.
     *
     * @param pfnpat PFN pattern to search for.
     *
     * @return str2_list List of
     *   {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
     *   objects containing matching lfn,pfn mappings.
     */
    public ArrayList getLFNWC(String pfnpat) throws RLSException {
      return lrcgetlfnwc(handle, pfnpat, null);
    }

    /**
     * Return PFNs mapped to LFN in the LRC database.
     *
     * @param lfn LFN to search for.
     *
     * @param offlim Offset and limit used to retrieve results incrementally.
     *	Use null, or 0,0, to retrieve all results.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2} objects
     *    containing matching lfn,pfn mappings.
     */
    public ArrayList getPFN(String lfn, RLSOffsetLimit offlim)
	  throws RLSException {
      return lrcgetpfn(handle, lfn, offlim);
    }

    /**
     * Bulk query of LFNs.
     *
     * @param strlist List of LFNs to search for. Elements of type
     *   <code>String</code>.
     *
     * @return List of query results.  ArrayList elements will be of
     *   type RLSString2Bulk.
     */
    public ArrayList getPFNBulk(ArrayList strlist) throws RLSException {
      int p = 0;
      
      Iterator iter = strlist.iterator();
      while (iter.hasNext())
	p = makestrlist(p, (String) iter.next());
      return lrcgetpfnbulk(handle, p);
    }

    /**
     * Return all PFNs mapped to LFN in the LRC database.
     *
     * @param lfn LFN to search for.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
     *    objects containing matching lfn,pfn mappings.
     */
    public ArrayList getPFN(String lfn)
	  throws RLSException {
      return lrcgetpfn(handle, lfn, null);
    }

    /**
     * Return PFNs mapped to LFNs matching lfnpat in the LRC database.
     *
     * @param lfnpat LFN pattern to search for.
     *
     * @param offlim Offset and limit used to retrieve results incrementally.
     *	Use null, or 0,0, to retrieve all results.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
     *    objects containing matching lfn,pfn mappings.
     */
    public ArrayList getPFNWC(String lfnpat, RLSOffsetLimit offlim)
	  throws RLSException {
      return lrcgetpfnwc(handle, lfnpat, offlim);
    }

    /**
     * Return all PFNs mapped to LFNs matching lfnpat in the LRC database.
     *
     * @param lfnpat LFN pattern to search for.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2}
     *    objects containing matching lfn,pfn mappings.
     */
    public ArrayList getPFNWC(String lfnpat)
	  throws RLSException {
      return lrcgetpfnwc(handle, lfnpat, null);
    }

    /**
     * Test if mapping exists in LRC.
     *
     * @param lfn Name of LFN.
     *
     * @param pfn Name of PFN.
     *
     * @return boolean True if mapping exists, else false.
     */
    public boolean mappingExists(String lfn, String pfn)
	  throws RLSException {
      return lrcmappingexists(handle, lfn, pfn);
    }

    /**
     * Rename an LFN.
     *
     * @param name Name of existing LFN.
     *
     * @param newname New name for LFN. Name must not exist.
     */
    public void renameLFN(String name, String newname)
	  throws RLSException {
      lrcrenamelfn(handle, name, newname);
    }

    /**
     * Bulk rename LFNs.
     *
     * @param str2list List of LFNs to rename. Elements of type 
     *   <code>RLSString2</code>, where <code>s1</code> is the name 
     *   of an existing LFN, and <code>s2</code> is the new name for
     *   the LFN.
     *
     * @return List of failed updates. ArrayList elements will be of
     *   type <code>RLSString2Bulk</code>.
     */
    public ArrayList renameLFNBulk(ArrayList str2list) throws RLSException {
      int p = 0;
      
      Iterator iter = str2list.iterator();
      while (iter.hasNext()) {
	RLSString2 str2 = (RLSString2) iter.next();
	p = makestr2list(p, str2.s1, str2.s2);
      }
      return lrcrenamelfnbulk(handle, p);
    }
   
    /**
     * Rename an PFN.
     *
     * @param name Name of existing PFN.
     *
     * @param newname New name for PFN. Name must not exist.
     */
    public void renamePFN(String name, String newname)
	  throws RLSException {
      lrcrenamepfn(handle, name, newname);
    }

    /**
     * Bulk rename PFNs.
     *
     * @param str2list List of PFNs to rename. Elements of type 
     *   <code>RLSString2</code>, where <code>s1</code> is the name 
     *   of an existing PFN, and <code>s2</code> is the new name for
     *   the PFN.
     *
     * @return List of failed updates. ArrayList elements will be of
     *   type <code>RLSString2Bulk</code>.
     */
    public ArrayList renamePFNBulk(ArrayList str2list) throws RLSException {
      int p = 0;
      
      Iterator iter = str2list.iterator();
      while (iter.hasNext()) {
	RLSString2 str2 = (RLSString2) iter.next();
	p = makestr2list(p, str2.s1, str2.s2);
      }
      return lrcrenamepfnbulk(handle, p);
    }
   
    /**
     * LRC servers send information about LFNs in their database to the
     * the list of RLI servers in the database, added with the following
     * function. Updates may be partitioned amongst multiple RLIs by
     * specifying one or more patterns for an RLI.
     *
     * @param rliurl URL of RLI server that LRC should send updates to.
     *
     * @param flags Should be zero or
     *    {@link org.globus.replica.rls.RLSClient#RLIFLAG_BLOOMFILTER RLSClient.RLIFLAG_BLOOMFILTER}
     *    if the RLI should be updated via Bloom filters.
     *
     * @param pattern If not NULL used to filter which LFNs are sent to rli_url.
     *	Standard Unix wildcard characters (*, ?) may be used to do wildcard
     *	matches.  Patterns are ignored if Bloom filters are used for updates.
     */
    public void rliAdd(String rliurl, int flags, String pattern)
	  throws RLSException {
      lrcrliadd(handle, rliurl, flags, pattern);
    }

    /**
     * Delete an entry from the LRC rli/partition tables.
     *
     * @param rliurl URL of RLI server to remove from LRC partition table.
     *
     * @param pattern If not null then only the specific rli_url/pattern is
     * removed, else all partition information for rli_url is removed.
     */
    public void rliDelete(String rliurl, String pattern)
	  throws RLSException {
      lrcrlidelete(handle, rliurl, pattern);
    }

    /**
     * Get RLI update partitions from LRC server.
     *
     * @param rliurl If not NULL identifies RLI that partition data will be
     *   retrieved for.  If null then all RLIs are retrieved.
     *
     * @param pattern If not NULL returns only partitions with matching patterns,
     *	otherwise all patterns are retrieved.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2} objects,
     *    containing the RLI URLs and patterns (if any) updated by the
     *    LRC.
     */
    public ArrayList rliGetPart(String rliurl, String pattern)
	  throws RLSException {
      return lrcrligetpart(handle, rliurl, pattern);
    }

    /**
     * Get info about RLI server updated by an LRC server.
     *
     * @param rliurl URL of RLI server to retrieve info for.
     *
     * @return info Data about the RLI server will be returned in an
     *   {@link org.globus.replica.rls.RLSRLIInfo RLSRLIInfo} object.
     */
    public RLSRLIInfo rliInfo(String rliurl) throws RLSException {
      return lrcrliinfo(handle, rliurl);
    }

    /**
     * Return URLs of RLIs that LRC sends updates to.
     *
     * @return A list of RLSRLIInfo objects containing the RLI URLs updated by
     *   this LRC.
     */
    public ArrayList rliList() throws RLSException {
      return lrcrlilist(handle);
    }

  }

  /**
   * This class implements the RLI interface to an RLS server.
   * @version 2.0.3
   */
  public class RLI {

    /**
     * Return LRCs mapped to LFN in the LRC database.
     *
     * @param lfn LFN to search for.
     *
     * @param offlim Offset and limit used to retrieve results incrementally.
     *	Use null, or 0,0, to retrieve all results.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2} objects
     *    containing matching lfn,lrc mappings.
     */
    public ArrayList getLRC(String lfn, RLSOffsetLimit offlim)
	  throws RLSException {
      return rligetlrc(handle, lfn, offlim);
    }

    /**
     * Return all LRCs mapped to LFNs matching lfnpat in the LRC database.
     *
     * @param lfnpat LFN pattern to search for.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2} objects
     *    containing matching lfn,lrc mappings.
     */
    public ArrayList getLRC(String lfn)
	  throws RLSException {
      return rligetlrc(handle, lfn, null);
    }

    /**
     * Bulk query of LFNs.
     *
     * @param strlist List of LFNs to search for. Elements of type
     *   <code>String</code>.
     *
     * @return List of query results.  ArrayList elements will be of
     *   type RLSString2Bulk.
     */
    public ArrayList getLRCBulk(ArrayList strlist) throws RLSException {
      int p = 0;
      
      Iterator iter = strlist.iterator();
      while (iter.hasNext())
	p = makestrlist(p, (String) iter.next());
      return rligetlrcbulk(handle, p);
    }

    /**
     * Return LRCs mapped to LFNs matching lfnpat in the LRC database.
     *
     * @param lfnpat LFN pattern to search for.
     *
     * @param offlim Offset and limit used to retrieve results incrementally.
     *	Use null, or 0,0, to retrieve all results.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2} objects
     *    containing matching lfn,lrc mappings.
     */
    public ArrayList getLRCWC(String lfnpat, RLSOffsetLimit offlim)
	  throws RLSException {
      return rligetlrcwc(handle, lfnpat, offlim);
    }

    /**
     * Return all LRCs mapped to lfns matching lfnpat in the LRC database.
     *
     * @param lfnpat LFN pattern to search for.
     *
     * @return str2_list List of {@link org.globus.replica.rls.RLSString2 rls.RLSString2} objects
     *    containing matching lfn,lrc mappings.
     */
    public ArrayList getLRCWC(String lfnpat)
	  throws RLSException {
      return rligetlrcwc(handle, lfnpat, null);
    }

    /**
     * Test if object (LFN or LRC) exists.
     *
     * @param key Name of object.
     *
     * @param objtype Type of object.
     *
     * @return boolean True if object exists, else false.
     */
    public boolean exists(String key, int objtype)
	  throws RLSException {
      return rliexists(handle, key, objtype);
    }

    /**
     * Bulk test if object (LFN or LRC) exists.
     *
     * @param strlist List of objects (LFN or LRC) to test for.
     *   Elements of type <code>String</code>.
     *
     * @param objtype Type of object.
     *
     * @return List of query results.  ArrayList elements will be of
     *   type <code>RLSString2Bulk</code>.
     */
    public ArrayList existsBulk(ArrayList strlist, int objtype) throws RLSException {
      int p = 0;
      
      Iterator iter = strlist.iterator();
      while (iter.hasNext())
	p = makestrlist(p, (String) iter.next());
      return rliexistsbulk(handle, p, objtype);
    }
    
    /**
     * Return URLs of LRCs that update this RLI.
     *
     * @return A list of RLSLRCInfo objects containing the RLI URLs updated by
     *   this LRC.
     */
    public ArrayList lrcList()
	  throws RLSException {
      return rlisenderlist(handle);
    }

    /**
     * Test if mapping exists in RLI.
     *
     * @param lfn Name of LFN.
     *
     * @param lrc Name of LRC.
     *
     * @return boolean True if mapping exists, else false.
     */
    public boolean mappingExists(String lfn, String lrc)
	  throws RLSException {
      return rlimappingexists(handle, lfn, lrc);
    }

    /**
     * Return URLs of RLIs that RLI sends updates to.
     *
     * @return A list of RLSRLIInfo objects containing the RLI URLs updated by
     *   this RLI.
     */
    public ArrayList rliList() throws RLSException {
      return rlirlilist(handle);
    }

  }

}
