#!/bin/bash
# 
# Copyright 1999-2006 University of Chicago
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 

server=${GLOBUS_RLS_SERVER:-'rls://localhost'}
client=${GLOBUS_RLS_CLI:-'globus-rls-cli'}
ncollect=20

function usage()
{
cat <<EOF
Usage: `basename $0` [options] COMMAND [file ...]
    COMMAND := add
             | delete
             | (exists|query) lrc (lfn|pfn)
             | (exists|query) rli lfn
             | rename (lfn|pfn)
    Options:
      -s SERVER, --server=SERVER
                RLS server (\$GLOBUS_RLS_SERVER, rls://localhost)
      -c CERT, --cert=CERT
                X509 cert (\$X509_USER_CERT, ~/.globus/usercert.pem)
      -k KEY, --key=KEY
                X509 key (\$X509_USER_KEY, ~/.globus/userkey.pem)
      -w, --wildcard
                enable wildcard query
      -t TIMEOUT, --timeout=TIMEOUT
                Timeout for RLS server connection in seconds (30)
      -p PATH, --path-to-cli=PATH
                RLS cmd line client (\$GLOBUS_RLS_CLI, globus-rls-cli)
      -n N, --ncollect=N
                Number of args to pass down to bulk command
      -h, --help
                This help message

    If no file is specified, input is taken from stdin.

    Note that 'add' works differently than globus-rls-cli in that it
    will add a mapping whether the LFN already exists or not.
EOF
}
TEMP=`getopt -o ht:s:p:c:k:n:w \
             --long help,timeout:,server:,client:,cert:,key:ncollect:wildcard \
             -n $0 -- "$@"`
if [ $? != 0 ] ; then usage >&2 ; exit 1 ; fi
eval set -- "$TEMP"

while true ; do
    case "$1" in
        -h|--help) usage; exit 0;;
        -s|--server) server=$2; shift 2 ;;
        -t|--timeout) timeout="-t $2"; shift 2 ;;
        -p|--program) client=$2; shift 2 ;;
        -c|--cert) export X509_USER_CERT="$2"; shift 2;;
        -k|--key)  export X509_USER_KEY="$2"; shift 2;;
        -w|--wildcard) wildcard=wildcard; shift;;
        -n|--ncollect) ncollect=$2; shift 2;;
        --) shift ; break ;;
        *) echo "Internal error"; exit 1 ;;
    esac
done

rls_cli="$client $timeout $server"

case "$1" in
    add|delete)
            cmd="bulk $1"; shift;;

    rename)
            if [ "$2" = "lfn" -o "$2" = "pfn" ]
            then cmd="bulk $1 $2"; shift 2
            else 
              echo "Error near '"$2"'" >&2
              usage >&2; exit 1
            fi;;

    exists|query)
            if [ "$2 $3" != "rli lfn" -a \
                 "$2 $3" != "lrc lfn" -a \
                 "$2 $3" != "lrc pfn" ]
            then
              echo "Error near '"$2 $3"'" >&2
              usage >&2; exit 1
            fi

            if [ $1 = "query" ]; then
                if [ -n "$wildcard" ]; then
                  cmd="$1 wildcard $2 $3"
                  ncollect=1
                else
                  cmd="bulk $1 $2 $3"
                fi
            else
              cmd="bulk $1 $2 $3 showall"
            fi
            shift 3;;

    *)
            if [ -n "$1" ]
            then
              echo "\"$1\": not a valid command"
            fi
            usage >&2; exit 1;;
esac

function gatherargs ()
{
  n=0
  line=""
  glob=0
  read args
  while [ -n "$args" ]; do
    let "n=(n+1)%${ncollect}"
    line="$line $args"
    if [ "$n" == 0 ]; then
        echo $* $line
        line=""
    fi
    read args
  done
  if [ -n "$line" ]; then
    echo $* $line
  fi
}

function run ()
{
  echo erf: $2
  if [ "$2" = "add" ]; then
    gatherargs "bulk create" | $rls_cli | sed -e 's/ *\(.*\),\(.*\):.*/\1 \2/' | gatherargs "bulk add" | $rls_cli
  else
    gatherargs $* | $rls_cli
  fi
}

# main
if [ -z "$1" ]; then
  run $cmd
else
  for file do
    if [ -r "$file" ]; then
      run "$cmd" <$file
    else
      echo "Cannot read file: \"$file\"" >&2
    fi
  done
fi
